from inspect import getattr_static
from pyteal import (
    CompileOptions,
    TealInputError,
    TealType,
    Tmpl,
    Expr,
    MAX_TEAL_VERSION,
    SubroutineDefinition,
    ABIReturnSubroutine,
    Seq,
    compileTeal,
    Reject,
    Mode,
    ScratchVar,
)
from beaker.decorators import get_handler_config


class TemplateVariable(Expr):
    """
    A Template Variable to be used as an attribute on LogicSignatures that
    need some hardcoded well defined behavior.

    If no ``name`` is supplied, the attribute name it was assigned to is used.

    """

    def __init__(self, stack_type: TealType, name: str = None):
        """initialize the TemplateVariable and the scratch var it is stored in"""
        assert stack_type in [TealType.bytes, TealType.uint64], "Must be bytes or uint"

        self.stack_type = stack_type
        self.scratch = ScratchVar(stack_type)
        self.name = name

    def get_name(self) -> str:
        """returns the name of the template variable that should be present in the output TEAL"""
        assert self.name is not None, TealInputError(
            "Name undefined in template variable"
        )
        return f"TMPL_{self.name.upper()}"

    def __str__(self) -> str:
        return f"(TemplateVariable {self.name})"

    def __teal__(self, options: CompileOptions):
        return self.scratch.load().__teal__(options)

    def has_return(self):
        """"""
        return False

    def type_of(self):
        """"""
        return self.stack_type

    def _init_expr(self) -> Expr:
        if self.stack_type is TealType.bytes:
            return self.scratch.store(Tmpl.Bytes(self.get_name()))
        return self.scratch.store(Tmpl.Int(self.get_name()))


class LogicSignature:
    """
    LogicSignature allows the definition of a logic signature program.

    A LogicSignature may include constants, subroutines, and :ref:TemplateVariables as attributes

    The `evauluate` method is the entry point to the application and must be overridden in any subclass
    to call the necessary logic.
    """

    def __init__(self, version: int = MAX_TEAL_VERSION):
        """initialize the logic signature and identify relevant attributes"""

        self.teal_version = version

        self.attrs = {
            m: (getattr(self, m), getattr_static(self, m))
            for m in sorted(list(set(dir(self.__class__)) - set(dir(super()))))
            if not m.startswith("__")
        }

        self.methods: dict[str, SubroutineDefinition] = {}

        self.template_variables: list[TemplateVariable] = []

        for name, (bound_attr, static_attr) in self.attrs.items():

            # Check for externals and internal methods
            handler_config = get_handler_config(bound_attr)

            if isinstance(static_attr, TemplateVariable):
                if static_attr.name is None:
                    static_attr.name = name
                self.template_variables.append(static_attr)

            elif handler_config.method_spec is not None:
                abi_meth = ABIReturnSubroutine(static_attr)
                if handler_config.referenced_self:
                    abi_meth.subroutine.implementation = bound_attr

                self.methods[name] = abi_meth.subroutine

            elif handler_config.subroutine is not None:
                if handler_config.referenced_self:
                    setattr(self, name, handler_config.subroutine(bound_attr))
                else:
                    setattr(
                        self.__class__,
                        name,
                        handler_config.subroutine(static_attr),
                    )

        template_expressions: list[Expr] = [
            tv._init_expr() for tv in self.template_variables
        ]

        self.program = compileTeal(
            Seq(*template_expressions, self.evaluate()),
            mode=Mode.Signature,
            version=self.teal_version,
            assembleConstants=True,
        )

    def evaluate(self):
        """
        evaluate is the main entry point to the logic of the lsig.

        Override this method to handle arbitrary logic.

        """
        return Reject()



from abc import abstractmethod, ABC
from copy import copy
from typing import Mapping, cast, Any, Optional
from algosdk.future.transaction import StateSchema
from pyteal import (
    abi,
    SubroutineFnWrapper,
    TealType,
    TealTypeError,
    Expr,
    CompileOptions,
    App,
    Bytes,
    Int,
    TealInputError,
    Assert,
    Not,
    MaybeValue,
    Txn,
    Seq,
    If,
)
from beaker.lib.storage import LocalBlob
from beaker.consts import MAX_GLOBAL_STATE, MAX_LOCAL_STATE
from beaker.lib.storage.global_blob import GlobalBlob


def get_default_for_type(stack_type, default):
    if default is not None:
        return default

    if stack_type == TealType.bytes:
        return Bytes("")
    else:
        return Int(0)


def stack_type_to_string(st: TealType):
    if st == TealType.uint64:
        return "uint64"
    if st == TealType.bytes:
        return "bytes"
    else:
        raise Exception("Only uint64 and bytes supported")


class StateValue(Expr):
    """Base Class for state values

    Attributes:
        stack_type: The type of the state value (either TealType.bytes or TealType.uint64)
        key: key to use to store the the value, default is name of class variable
        default: Default value for the state value
        static: Boolean flag to denote that this state value can only be set once and not deleted.
        descr: Description of the state value to provide some information to clients
    """

    def __init__(
        self,
        stack_type: TealType,
        key: Expr = None,
        default: Expr = None,
        static: bool = False,
        descr: str = None,
    ):
        super().__init__()

        self.stack_type = stack_type
        self.static = static
        self.descr = descr

        if key is not None and key.type_of() != TealType.bytes:
            raise TealTypeError(key.type_of(), TealType.bytes)
        self.key = key

        if default is not None and default.type_of() != self.stack_type:
            raise TealTypeError(default.type_of(), self.stack_type)
        self.default = default

    # Required methods for `Expr subclass`
    def has_return(self) -> bool:
        return False

    def type_of(self) -> TealType:
        return self.stack_type

    def __teal__(self, options: "CompileOptions"):
        return self.get().__teal__(options)

    def __str__(self) -> str:
        return f"StateValue {self.key}"

    def str_key(self) -> str:
        """returns the string held by the key Bytes object"""
        return cast(Bytes, self.key).byte_str.replace('"', "")

    def increment(self, cnt: Expr = Int(1)) -> Expr:
        """helper to increment a counter"""
        check_is_int(self)
        check_not_static(self)

        return self.set(self.get() + cnt)

    def decrement(self, cnt: Expr = Int(1)) -> Expr:
        """helper to decrement a counter"""
        check_is_int(self)
        check_not_static(self)

        return self.set(self.get() - cnt)

    def set_default(self) -> Expr:
        """sets the default value if one is provided, if none provided sets the zero value for its type"""

        return self.set(get_default_for_type(self.stack_type, self.default))

    def is_default(self) -> Expr:
        """checks to see if the value set equals the default value"""
        default = get_default_for_type(self.stack_type, self.default)
        return self.get() == default

    @abstractmethod
    def set(self, val: Expr) -> Expr:
        """sets the value to the argument passed"""

    @abstractmethod
    def get(self) -> Expr:
        """gets the value stored for this state value"""

    @abstractmethod
    def get_maybe(self) -> MaybeValue:
        """gets a MaybeValue that can be used for existence check"""

    @abstractmethod
    def get_must(self) -> Expr:
        """gets the value stored at the key. if none is stored, Assert out of the program"""

    @abstractmethod
    def get_else(self, val: Expr) -> Expr:
        """gets the value stored at the key. if none is stored, return the value passed"""

    @abstractmethod
    def exists(self) -> Expr:
        """checks if the value is set (to whatever value. Returns Int(1) if value is set, Int(0) otherwise."""

    @abstractmethod
    def delete(self) -> Expr:
        """deletes the key from state, if the value is static it will be a compile time error"""


class ReservedStateValue(ABC):
    """Base Class for ReservedStateValues

    Attributes:
        stack_type (TealType): The type of the state value (either TealType.bytes or TealType.uint64)
        max_keys (int): Maximum number of keys to reserve for this reserved state value
        key_gen (subroutine): A subroutine returning TealType.bytes, used to create a key where some data is stored.
        descr (str): Description of the state value to provide some information to clients

    """

    def __init__(
        self,
        stack_type: TealType,
        max_keys: int,
        key_gen: SubroutineFnWrapper = None,
        descr: str = None,
    ):
        self.stack_type = stack_type
        self.max_keys = max_keys
        self.descr = descr

        if key_gen is not None and key_gen.type_of() != TealType.bytes:
            raise TealTypeError(key_gen.type_of(), TealType.bytes)

        self.key_generator = key_gen

    @abstractmethod
    def __getitem__(self, key_seed: Expr | abi.BaseType) -> StateValue:
        """Method to access the state value with the key seed provided"""


class ApplicationStateValue(StateValue):
    """Allows storage of global state values for an application

    Attributes:
        stack_type: The type of the state value (either TealType.bytes or TealType.uint64)
        key: key to use to store the the value, default is name of class variable
        default: Default value for the state value
        static: Boolean flag to denote that this state value can only be set once and not deleted.
        descr: Description of the state value to provide some information to clients
    """

    def __str__(self) -> str:
        return f"ApplicationStateValue {self.key}"

    def set(self, val: Expr) -> Expr:
        check_match_type(self, val)

        if self.key is None:
            raise TealInputError(f"ApplicationStateValue {self} has no key defined")

        if self.static:
            return Seq(
                v := App.globalGetEx(Int(0), self.key),
                Assert(Not(v.hasValue())),
                App.globalPut(self.key, val),
            )

        return App.globalPut(self.key, val)

    def get(self) -> Expr:
        if self.key is None:
            raise TealInputError(f"ApplicationStateValue {self} has no key defined")

        return App.globalGet(self.key)

    def get_maybe(self) -> MaybeValue:
        if self.key is None:
            raise TealInputError(f"ApplicationStateValue {self} has no key defined")

        return App.globalGetEx(Int(0), self.key)

    def get_must(self) -> Expr:
        if self.key is None:
            raise TealInputError(f"ApplicationStateValue {self} has no key defined")

        return Seq(val := self.get_maybe(), Assert(val.hasValue()), val.value())

    def get_else(self, val: Expr) -> Expr:
        check_match_type(self, val)

        if self.key is None:
            raise TealInputError(f"ApplicationStateValue {self} has no key defined")

        return If((v := App.globalGetEx(Int(0), self.key)).hasValue(), v.value(), val)

    def exists(self) -> Expr:
        return Seq(val := self.get_maybe(), val.hasValue())

    def delete(self) -> Expr:
        check_not_static(self)

        if self.key is None:
            raise TealInputError(f"ApplicationStateValue {self} has no key defined")

        return App.globalDel(self.key)


class ReservedApplicationStateValue(ReservedStateValue):
    """Reservedally keyed Application State

    Attributes:
        stack_type (TealType): The type of the state value (either TealType.bytes or TealType.uint64)
        max_keys (int): Maximum number of keys to reserve for this reserved state value
        key_gen (SubroutineFnWrapper): A subroutine returning TealType.bytes, used to create a key where some data is stored.
        descr (str): Description of the state value to provide some information to clients
    """

    def __init__(
        self,
        stack_type: TealType,
        max_keys: int,
        key_gen: SubroutineFnWrapper = None,
        descr: str = None,
    ):
        super().__init__(stack_type, max_keys, key_gen, descr)

        if max_keys <= 0 or max_keys > MAX_GLOBAL_STATE:
            raise Exception(f"max keys expected to be between 0 and {MAX_GLOBAL_STATE}")

    def __getitem__(self, key_seed: Expr | abi.BaseType) -> ApplicationStateValue:
        """Method to access the state value with the key seed provided"""
        key = key_seed

        if isinstance(key_seed, abi.BaseType):
            key = key_seed.encode()

        key = cast(Expr, key)

        if self.key_generator is not None:
            key = self.key_generator(key)

        return ApplicationStateValue(
            stack_type=self.stack_type, key=key, descr=self.descr
        )


class AccountStateValue(StateValue):
    """Allows storage of global state values for an account opted into an application

    Attributes:
        stack_type: The type of the state value (either TealType.bytes or TealType.uint64)
        key: key to use to store the the value, default is name of class variable
        default: Default value for the state value
        static: Boolean flag to denote that this state value can only be set once and not deleted.
        descr: Description of the state value to provide some information to clients
    """

    def __init__(
        self,
        stack_type: TealType,
        key: Expr = None,
        default: Expr = None,
        static: bool = False,
        descr: str = None,
    ):
        super().__init__(stack_type, key, default, static, descr)
        self.acct: Expr = Txn.sender()

    def __str__(self) -> str:
        return f"AccountStateValue {self.acct} {self.key}"

    def set(self, val: Expr) -> Expr:
        check_match_type(self, val)

        if self.key is None:
            raise TealInputError(f"AccountStateValue {self} has no key defined")

        if self.acct is None:
            raise TealInputError(f"AccountStateValue {self} has no account defined")

        if self.static:
            return Seq(
                v := self.get_maybe(),
                Assert(Not(v.hasValue())),
                App.localPut(self.acct, self.key, val),
            )

        return App.localPut(self.acct, self.key, val)

    def get(self) -> Expr:
        if self.key is None:
            raise TealInputError(f"AccountStateValue {self} has no key defined")

        if self.acct is None:
            raise TealInputError(f"AccountStateValue {self} has no account defined")

        return App.localGet(self.acct, self.key)

    def get_maybe(self) -> MaybeValue:
        if self.key is None:
            raise TealInputError(f"AccountStateValue {self} has no key defined")

        if self.acct is None:
            raise TealInputError(f"AccountStateValue {self} has no account defined")

        return App.localGetEx(self.acct, Int(0), self.key)

    def get_must(self) -> Expr:
        if self.key is None:
            raise TealInputError(f"AccountStateValue {self} has no key defined")
        if self.acct is None:
            raise TealInputError(f"AccountStateValue {self} has no account defined")

        return Seq(val := self.get_maybe(), Assert(val.hasValue()), val.value())

    def get_else(self, val: Expr) -> Expr:
        check_match_type(self, val)

        if self.key is None:
            raise TealInputError(f"AccountStateValue {self} has no key defined")
        if self.acct is None:
            raise TealInputError(f"AccountStateValue {self} has no account defined")

        return If(
            (v := App.localGetEx(self.acct, Int(0), self.key)).hasValue(),
            v.value(),
            val,
        )

    def exists(self) -> Expr:
        return Seq(val := self.get_maybe(), val.hasValue())

    def delete(self) -> Expr:
        if self.key is None:
            raise TealInputError(f"AccountStateValue {self} has no key defined")
        if self.acct is None:
            raise TealInputError(f"AccountStateValue {self} has no account defined")

        return App.localDel(self.acct, self.key)

    def __getitem__(self, acct: Expr) -> "AccountStateValue":
        asv = copy(self)
        asv.acct = acct
        return asv


class ReservedAccountStateValue(ReservedStateValue):
    """Reservedally keyed Account State

    Attributes:
        stack_type (TealType): The type of the state value (either TealType.bytes or TealType.uint64)
        max_keys (int): Maximum number of keys to reserve for this reserved state value
        key_gen (SubroutineFnWrapper): A subroutine returning TealType.bytes, used to create a key where some data is stored.
        descr (str): Description of the state value to provide some information to clients
    """

    def __init__(
        self,
        stack_type: TealType,
        max_keys: int,
        key_gen: SubroutineFnWrapper = None,
        descr: str = None,
    ):
        super().__init__(stack_type, max_keys, key_gen, descr)

        if max_keys <= 0 or max_keys > MAX_LOCAL_STATE:
            raise Exception(f"max keys expected to be between 0 and {MAX_LOCAL_STATE}")

    def __getitem__(self, key_seed: Expr | abi.BaseType) -> AccountStateValue:
        """Access AccountState value given key_seed"""
        key = key_seed

        if isinstance(key_seed, abi.BaseType):
            key = key_seed.encode()

        if self.key_generator is not None:
            key = self.key_generator(key)

        return AccountStateValue(stack_type=self.stack_type, key=cast(Expr, key))


class StateBlob(ABC):
    def __init__(self, num_keys: int):
        self.num_keys = num_keys

    @abstractmethod
    def initialize(self) -> Expr:
        ...

    @abstractmethod
    def read(self, start: Expr, stop: Expr) -> Expr:
        """
        Reads some bytes from the buffer

        Args:
            start: An ``Expr`` that represents the start index to read from. Should evaluate to ``uint64``.
            stop: An ``Expr`` that represents the stop index to read until. Should evaluate to ``uint64``.
        Returns:
            The bytes read from the blob from start to stop
        """
        ...

    @abstractmethod
    def write(self, start: Expr, buff: Expr) -> Expr:
        """
        Writes the buffer to the blob

        Args:
            start: An ``Expr`` that represents where to start writing. Should evaluate to ``uint64``.
            buff: An ``Expr`` that represents the bytes to write. Should evaluate to ``bytes``.

        """
        ...

    @abstractmethod
    def read_byte(self, idx: Expr) -> Expr:
        """
        Reads a single byte from the given index

        Args:
            idx: An ``Expr`` that represents the index into the blob to read the byte from. Should evaluate to ``uint64``.

        Returns:
            A single byte as a ``uint64``

        """
        ...

    @abstractmethod
    def write_byte(self, idx: Expr, byte: Expr) -> Expr:
        """
        Writes a single byte to the given index

        Args:
            idx: An ``Expr`` that represents the index to write the byte to. Should evaluate to ``uint64``.
            byte: An ``Expr`` That represents the index to write the byte to. Should evaluate to ``uint64``.

        """
        ...


class AccountStateBlob(StateBlob):
    def __init__(self, keys: Optional[int | list[int]] = None):
        self.blob = LocalBlob(keys=keys)
        self.acct: Expr = Txn.sender()

        super().__init__(self.blob._max_keys)

    def __getitem__(self, acct: Expr) -> "AccountStateBlob":
        asv = copy(self)
        asv.acct = acct
        return asv

    def initialize(self) -> Expr:
        return self.blob.zero(acct=self.acct)

    def write(self, start: Expr, buff: Expr) -> Expr:
        return self.blob.write(start, buff, acct=self.acct)

    def read(self, start: Expr, stop: Expr) -> Expr:
        return self.blob.read(start, stop, acct=self.acct)

    def read_byte(self, idx: Expr) -> Expr:
        return self.blob.get_byte(idx, acct=self.acct)

    def write_byte(self, idx: Expr, byte: Expr) -> Expr:
        return self.blob.set_byte(idx, byte, acct=self.acct)


class ApplicationStateBlob(StateBlob):
    def __init__(self, keys: Optional[int | list[int]] = None):
        self.blob = GlobalBlob(keys=keys)
        super().__init__(self.blob._max_keys)

    def initialize(self) -> Expr:
        return self.blob.zero()

    def write(self, start: Expr, buff: Expr) -> Expr:
        return self.blob.write(start, buff)

    def read(self, start: Expr, stop: Expr) -> Expr:
        return self.blob.read(start, stop)

    def read_byte(self, idx: Expr) -> Expr:
        return self.blob.get_byte(idx)

    def write_byte(self, idx: Expr, byte: Expr) -> Expr:
        return self.blob.set_byte(idx, byte)


def check_not_static(sv: StateValue):
    if sv.static:
        raise TealInputError(f"StateValue {sv} is static")


def check_is_int(sv: StateValue):
    if sv.stack_type != TealType.uint64:
        raise TealInputError(f"StateValue {sv} is not integer type")


def check_match_type(sv: StateValue, val: Expr):
    in_type = val.type_of()
    if in_type != sv.stack_type and in_type != TealType.anytype:
        raise TealTypeError(in_type, sv.stack_type)


class State:
    """holds all the declared and reserved state values for this storage type"""

    def __init__(
        self, fields: Mapping[str, StateValue | ReservedStateValue | StateBlob]
    ):
        self.declared_vals: dict[str, StateValue] = {
            k: v for k, v in fields.items() if isinstance(v, StateValue)
        }
        self.__dict__.update(self.declared_vals)

        self.blob_vals: dict[str, StateBlob] = {
            k: v for k, v in fields.items() if isinstance(v, StateBlob)
        }
        self.__dict__.update(self.blob_vals)

        self.reserved_vals: dict[str, ReservedStateValue] = {
            k: v for k, v in fields.items() if isinstance(v, ReservedStateValue)
        }
        self.__dict__.update(self.reserved_vals)

        self.num_uints = len(
            [l for l in self.declared_vals.values() if l.stack_type == TealType.uint64]
        ) + sum(
            [
                l.max_keys
                for l in self.reserved_vals.values()
                if l.stack_type == TealType.uint64
            ]
        )

        self.num_byte_slices = (
            len(
                [
                    l
                    for l in self.declared_vals.values()
                    if l.stack_type == TealType.bytes
                ]
            )
            + sum(
                [
                    l.max_keys
                    for l in self.reserved_vals.values()
                    if l.stack_type == TealType.bytes
                ]
            )
            + sum([b.num_keys for b in self.blob_vals.values()])
        )

    def dictify(self) -> dict[str, dict[str, Any]]:
        """Convert the state to a dict for encoding"""
        return {
            "declared": {
                k: {
                    "type": stack_type_to_string(v.stack_type),
                    "key": v.str_key(),
                    "descr": v.descr if v.descr is not None else "",
                }
                for k, v in self.declared_vals.items()
            },
            "reserved": {
                k: {
                    "type": stack_type_to_string(v.stack_type),
                    "max_keys": v.max_keys,
                    "descr": v.descr if v.descr is not None else "",
                }
                for k, v in self.reserved_vals.items()
            },
        }

    def schema(self) -> StateSchema:
        """gets the schema as num uints/bytes for app create transactions"""
        return StateSchema(
            num_uints=self.num_uints, num_byte_slices=self.num_byte_slices
        )


class ApplicationState(State):
    def __init__(
        self,
        fields: Mapping[
            str,
            ApplicationStateValue
            | ReservedApplicationStateValue
            | ApplicationStateBlob,
        ],
    ):
        super().__init__(fields)
        if (total := self.num_uints + self.num_byte_slices) > MAX_GLOBAL_STATE:
            raise Exception(
                f"Too much application state, expected {total} <= {MAX_GLOBAL_STATE}"
            )

    def initialize(self) -> Expr:
        """Generate expression from state values to initialize a default value"""
        return Seq(
            *[
                v.set_default()
                for v in self.declared_vals.values()
                if not v.static or (v.static and v.default is not None)
            ]
            + [v.initialize() for v in self.blob_vals.values()]
        )


class AccountState(State):
    def __init__(
        self,
        fields: Mapping[
            str, AccountStateValue | ReservedAccountStateValue | AccountStateBlob
        ],
    ):
        super().__init__(fields)
        if (total := self.num_uints + self.num_byte_slices) > MAX_LOCAL_STATE:
            raise Exception(
                f"Too much account state, expected {total} <= {MAX_LOCAL_STATE}"
            )

    def initialize(self, acct: Expr = Txn.sender()) -> Expr:
        """Generate expression from state values to initialize a default value"""
        return Seq(
            *[
                cast(AccountStateValue, v)[acct].set_default()
                for v in self.declared_vals.values()
                if not v.static or (v.static and v.default is not None)
            ]
            + [v.initialize() for v in self.blob_vals.values()]
        )



from abc import ABC, abstractmethod
from typing import Optional
from pyteal import BytesZero, Int, Expr, Extract, Bytes, TealTypeError

blob_page_size = 128 - 1  # need 1 byte for key
BLOB_PAGE_SIZE = Int(blob_page_size)
EMPTY_PAGE = BytesZero(BLOB_PAGE_SIZE)


class Blob(ABC):
    """
    Blob is a class holding static methods to work with the global storage of an application as a binary large object

    """

    def __init__(self, key_limit: int, /, *, keys: Optional[int | list[int]] = None):

        _keys: list[int] = []

        if keys is None:
            _keys = [x for x in range(key_limit)]
        elif type(keys) is int:
            _keys = [x for x in range(keys)]
        elif type(keys) is list:
            _keys = keys
        else:
            raise TealTypeError(type(keys), int | list[int])

        assert max(_keys) <= 255, "larger than 1 byte key supplied"
        assert sorted(_keys) == _keys, "keys provided are not sorted"

        self.byte_keys = [key.to_bytes(1, "big") for key in _keys]
        self.byte_key_str = Bytes("base16", b"".join(self.byte_keys).hex())

        self.int_keys = [Int(key) for key in _keys]
        self.start_key = self.int_keys[0]

        self._max_keys = len(_keys)
        self._max_bytes = self._max_keys * blob_page_size
        self._max_bits = self._max_bytes * 8

        self.max_keys = Int(self._max_keys)
        self.max_bytes = Int(self._max_bytes)

    def _key(self, i) -> Expr:
        return Extract(self.byte_key_str, i, Int(1))

    def _key_idx(self, idx: Int) -> Expr:
        return idx / BLOB_PAGE_SIZE

    def _offset_for_idx(self, idx: Int) -> Expr:
        return idx % BLOB_PAGE_SIZE

    @abstractmethod
    def zero(self) -> Expr:
        ...

    @abstractmethod
    def get_byte(self, idx) -> Expr:
        ...

    @abstractmethod
    def set_byte(self, idx, byte) -> Expr:
        ...

    @abstractmethod
    def read(self, bstart, bstop) -> Expr:
        ...

    @abstractmethod
    def write(self, bstart, buff) -> Expr:
        ...



from typing import Optional
from pyteal import (
    Txn,
    App,
    Bytes,
    Concat,
    Expr,
    Extract,
    For,
    GetByte,
    If,
    Int,
    Len,
    Or,
    ScratchVar,
    Seq,
    SetByte,
    Subroutine,
    Substring,
    TealType,
)
from beaker.lib.storage.blob import BLOB_PAGE_SIZE, EMPTY_PAGE, Blob


class LocalBlob(Blob):
    """
    Blob is a class holding static methods to work with the local storage of an account as a binary large object

    The `zero` method must be called on an account on opt in and the schema of the local storage should be 16 bytes
    """

    def __init__(self, /, *, keys: Optional[int | list[int]] = None):
        super().__init__(16, keys=keys)

    def zero(self, acct: Expr = Txn.sender()) -> Expr:
        """
        initializes local state of an account to all zero bytes

        This allows us to be lazy later and _assume_ all the strings are the same size

        """

        @Subroutine(TealType.none)
        def _impl(acct):
            return Seq(
                *[App.localPut(acct, Bytes(bk), EMPTY_PAGE) for bk in self.byte_keys]
            )

        return _impl(acct)

    def get_byte(self, idx, acct: Expr = Txn.sender()):
        """
        Get a single byte from local storage of an account by index
        """

        @Subroutine(TealType.uint64)
        def _impl(acct, idx):
            return GetByte(
                App.localGet(acct, self._key(self._key_idx(idx))),
                self._offset_for_idx(idx),
            )

        return _impl(acct, idx)

    def set_byte(self, idx, byte, acct: Expr = Txn.sender()):
        """
        Set a single byte from local storage of an account by index
        """

        @Subroutine(TealType.none)
        def _impl(acct, idx, byte):
            return Seq(
                (key := ScratchVar()).store(self._key(self._key_idx(idx))),
                App.localPut(
                    acct,
                    key.load(),
                    SetByte(
                        App.localGet(acct, key.load()), self._offset_for_idx(idx), byte
                    ),
                ),
            )

        return _impl(acct, idx, byte)

    def read(self, bstart, bend, acct: Expr = Txn.sender()) -> Expr:
        """
        read bytes between bstart and bend from local storage of an account by index
        """

        @Subroutine(TealType.bytes)
        def _impl(acct, bstart, bend):
            start_key_idx = self._key_idx(bstart)
            start_offset = self._offset_for_idx(bstart)

            stop_key_idx = self._key_idx(bend)
            stop_offset = self._offset_for_idx(bend)

            key_idx = ScratchVar()
            buff = ScratchVar()

            start = ScratchVar()
            stop = ScratchVar()

            init = key_idx.store(start_key_idx)
            cond = key_idx.load() <= stop_key_idx
            incr = key_idx.store(key_idx.load() + Int(1))

            return Seq(
                buff.store(Bytes("")),
                For(init, cond, incr).Do(
                    Seq(
                        start.store(
                            If(key_idx.load() == start_key_idx, start_offset, Int(0))
                        ),
                        stop.store(
                            If(
                                key_idx.load() == stop_key_idx,
                                stop_offset,
                                BLOB_PAGE_SIZE,
                            )
                        ),
                        buff.store(
                            Concat(
                                buff.load(),
                                Substring(
                                    App.localGet(acct, self._key(key_idx.load())),
                                    start.load(),
                                    stop.load(),
                                ),
                            )
                        ),
                    )
                ),
                buff.load(),
            )

        return _impl(acct, bstart, bend)

    def write(self, bstart, buff, acct: Expr = Txn.sender()) -> Expr:
        """
        write bytes between bstart and len(buff) to local storage of an account
        """

        @Subroutine(TealType.none)
        def _impl(acct, bstart, buff):

            start_key_idx = self._key_idx(bstart)
            start_offset = self._offset_for_idx(bstart)

            stop_key_idx = self._key_idx(bstart + Len(buff))
            stop_offset = self._offset_for_idx(bstart + Len(buff))

            key_idx = ScratchVar()
            start = ScratchVar()
            stop = ScratchVar()
            written = ScratchVar()

            init = key_idx.store(start_key_idx)
            cond = key_idx.load() <= stop_key_idx
            incr = key_idx.store(key_idx.load() + Int(1))

            delta = ScratchVar()

            return Seq(
                written.store(Int(0)),
                For(init, cond, incr).Do(
                    Seq(
                        start.store(
                            If(key_idx.load() == start_key_idx, start_offset, Int(0))
                        ),
                        stop.store(
                            If(
                                key_idx.load() == stop_key_idx,
                                stop_offset,
                                BLOB_PAGE_SIZE,
                            )
                        ),
                        App.localPut(
                            acct,
                            self._key(key_idx.load()),
                            If(
                                Or(
                                    stop.load() != BLOB_PAGE_SIZE,
                                    start.load() != Int(0),
                                )
                            )  # Its a partial write
                            .Then(
                                Seq(
                                    delta.store(stop.load() - start.load()),
                                    Concat(
                                        Substring(
                                            App.localGet(
                                                acct, self._key(key_idx.load())
                                            ),
                                            Int(0),
                                            start.load(),
                                        ),
                                        Extract(buff, written.load(), delta.load()),
                                        Substring(
                                            App.localGet(
                                                acct, self._key(key_idx.load())
                                            ),
                                            stop.load(),
                                            BLOB_PAGE_SIZE,
                                        ),
                                    ),
                                )
                            )
                            .Else(
                                Seq(
                                    delta.store(BLOB_PAGE_SIZE),
                                    Extract(buff, written.load(), BLOB_PAGE_SIZE),
                                )
                            ),
                        ),
                        written.store(written.load() + delta.load()),
                    )
                ),
            )

        return _impl(acct, bstart, buff)



import pytest

from algosdk.account import generate_account
from algosdk.error import KMDHTTPError
from .kmd import delete_account, get_accounts, add_account


def test_get_accounts():
    accts = get_accounts()
    assert (
        len(accts) > 0
    ), "Should have accounts (make sure its a sandbox with default wallet settings)"


def test_add_remove_account():
    pk, addr = generate_account()
    addr_added = add_account(pk)
    assert addr == addr_added, "Expected added address to match generated address"

    with pytest.raises(KMDHTTPError):
        add_account("lol")

    delete_account(addr_added)

    with pytest.raises(KMDHTTPError):
        delete_account("lol")



from dataclasses import asdict, dataclass, field, replace, astuple
from enum import Enum
from functools import wraps
from inspect import get_annotations, signature, Parameter
from typing import Optional, Callable, Final, cast, Any, TypeVar
from types import FunctionType
from algosdk.abi import Method
from pyteal import (
    abi,
    ABIReturnSubroutine,
    And,
    Global,
    App,
    Assert,
    AssetHolding,
    BareCallActions,
    CallConfig,
    Expr,
    Int,
    MethodConfig,
    OnCompleteAction,
    Seq,
    Subroutine,
    SubroutineFnWrapper,
    TealInputError,
    TealType,
    TealTypeError,
    Bytes,
    Txn,
)

from beaker.state import AccountStateValue, ApplicationStateValue

HandlerFunc = Callable[..., Expr]

_handler_config_attr: Final[str] = "__handler_config__"


CheckExpr = Callable[..., Expr]
ABIType = TypeVar("ABIType", bound=abi.BaseType)


DefaultArgumentType = Expr | FunctionType | int | bytes | str


class DefaultArgumentClass(str, Enum):
    ABIMethod = "abi-method"
    LocalState = "local-state"
    GlobalState = "global-state"
    Constant = "constant"


class DefaultArgument:
    """DefaultArgument is a container for any arguments that may be resolved prior to calling some target method"""

    def __init__(
        self,
        resolver: DefaultArgumentType,
    ):
        self.resolver = resolver

        match resolver:

            # Expr types
            case AccountStateValue():
                self.resolvable_class = DefaultArgumentClass.LocalState
            case ApplicationStateValue():
                self.resolvable_class = DefaultArgumentClass.GlobalState
            case Bytes() | Int():
                self.resolvable_class = DefaultArgumentClass.Constant

            # Native types
            case int() | str() | bytes():
                self.resolvable_class = DefaultArgumentClass.Constant

            # FunctionType
            case _:
                # Fall through, if its not got a valid handler config, raise error
                hc = get_handler_config(cast(HandlerFunc, resolver))
                if hc.method_spec is None or not hc.read_only:
                    raise TealTypeError(self.resolver, DefaultArgumentType)

                self.resolvable_class = DefaultArgumentClass.ABIMethod

    def resolve_hint(self):
        match self.resolver:
            # Expr types
            case AccountStateValue() | ApplicationStateValue():
                return self.resolver.str_key()
            case Bytes():
                return self.resolver.byte_str.replace('"', "")
            case Int():
                return self.resolver.value
            # Native types
            case int() | bytes() | str():
                return self.resolver

            # FunctionType
            case _:
                # Fall through, if its not got a valid handler config, raise error
                hc = get_handler_config(cast(HandlerFunc, self.resolver))
                if hc.method_spec is None or not hc.read_only:
                    raise TealTypeError(self.resolver, DefaultArgumentType)

                return hc.method_spec.dictify()

    def dictify(self) -> dict[str, Any]:
        return {"source": self.resolvable_class.value, "data": self.resolve_hint()}


@dataclass
class HandlerConfig:
    """HandlerConfig contains all the extra bits of info about a given ABI method"""

    method_spec: Optional[Method] = field(kw_only=True, default=None)
    subroutine: Optional[Subroutine] = field(kw_only=True, default=None)
    bare_method: Optional[BareCallActions] = field(kw_only=True, default=None)

    referenced_self: bool = field(kw_only=True, default=False)
    structs: Optional[dict[str, abi.NamedTuple]] = field(kw_only=True, default=None)

    default_arguments: Optional[dict[str, DefaultArgument]] = field(
        kw_only=True, default=None
    )

    method_config: Optional[MethodConfig] = field(kw_only=True, default=None)
    read_only: bool = field(kw_only=True, default=False)

    def hints(self) -> "MethodHints":
        mh: dict[str, Any] = {"read_only": self.read_only}

        if (
            self.default_arguments is not None
            and len(self.default_arguments.keys()) > 0
        ):
            mh["default_arguments"] = self.default_arguments

        if self.structs is not None:
            structs: dict[str, dict[str, str | list[tuple[str, str]]]] = {}
            for arg_name, model_spec in self.structs.items():
                annos: list[tuple[str, Any]] = list(model_spec.__annotations__.items())
                structs[arg_name] = {
                    "name": str(model_spec.__name__),  # type: ignore[attr-defined]
                    "elements": [
                        (name, str(abi.algosdk_from_annotation(typ.__args__[0])))
                        for name, typ in annos
                    ],
                }

            mh["structs"] = structs

        return MethodHints(**mh)

    def is_create(self) -> bool:
        if self.method_config is None:
            return False

        return any(
            map(
                lambda cc: cc == CallConfig.CREATE or cc == CallConfig.ALL,
                astuple(self.method_config),
            )
        )

    def is_update(self) -> bool:
        return (
            self.method_config is not None
            and self.method_config.update_application != CallConfig.NEVER
        )

    def is_delete(self) -> bool:
        return (
            self.method_config is not None
            and self.method_config.delete_application != CallConfig.NEVER
        )

    def is_opt_in(self) -> bool:
        return (
            self.method_config is not None
            and self.method_config.opt_in != CallConfig.NEVER
        )

    def is_clear_state(self) -> bool:
        return (
            self.method_config is not None
            and self.method_config.clear_state != CallConfig.NEVER
        )

    def is_close_out(self) -> bool:
        return (
            self.method_config is not None
            and self.method_config.close_out != CallConfig.NEVER
        )


def get_handler_config(fn: HandlerFunc) -> HandlerConfig:
    if hasattr(fn, _handler_config_attr):
        return cast(HandlerConfig, getattr(fn, _handler_config_attr))
    return HandlerConfig()


def set_handler_config(fn: HandlerFunc, **kwargs):
    handler_config = get_handler_config(fn)
    setattr(fn, _handler_config_attr, replace(handler_config, **kwargs))


@dataclass
class MethodHints:
    """MethodHints provides hints to the caller about how to call the method"""

    #: hint to indicate this method can be called through Dryrun
    read_only: bool = field(kw_only=True, default=False)
    #: hint to provide names for tuple argument indices method_name=>param_name=>{name:str, elements:[str,str]}
    structs: Optional[dict[str, dict[str, str | list[tuple[str, str]]]]] = field(
        kw_only=True, default=None
    )
    #: defaults
    default_arguments: Optional[dict[str, DefaultArgument]] = field(
        kw_only=True, default=None
    )

    def empty(self) -> bool:
        return (
            self.structs is None
            and self.default_arguments is None
            and not self.read_only
        )

    def dictify(self) -> dict[str, Any]:
        d: dict[str, Any] = {}
        if self.read_only:
            d["read_only"] = True
        if self.default_arguments is not None:
            d["default_arguments"] = {
                k: v.dictify() for k, v in self.default_arguments.items()
            }
        if self.structs is not None:
            d["structs"] = self.structs
        return d


class Authorize:
    """Authorize contains methods that may be used as values to the `authorize` keyword of the `handle` decorator"""

    @staticmethod
    def only(addr: Expr):
        """require that the sender of the app call match exactly the address passed"""

        if addr.type_of() != TealType.bytes:
            raise TealTypeError(addr.type_of(), TealType.bytes)

        @Subroutine(TealType.uint64, name="auth_only")
        def _impl(sender: Expr) -> Expr:
            return sender == addr

        return _impl

    @staticmethod
    def holds_token(asset_id: Expr):
        """require that the sender of the app call holds >0 of the asset id passed"""

        if asset_id.type_of() != TealType.uint64:
            raise TealTypeError(asset_id.type_of(), TealType.uint64)

        @Subroutine(TealType.uint64, name="auth_holds_token")
        def _impl(sender: Expr):
            return Seq(
                bal := AssetHolding.balance(sender, asset_id),
                And(bal.hasValue(), bal.value() > Int(0)),
            )

        return _impl

    @staticmethod
    def opted_in(app_id: Expr = Global.current_application_id()):
        """require that the sender of the app call has already opted-in to a given app id"""

        if app_id.type_of() != TealType.uint64:
            raise TealTypeError(app_id.type_of(), TealType.uint64)

        @Subroutine(TealType.uint64, name="auth_opted_in")
        def _impl(sender: Expr):
            return App.optedIn(sender, app_id)

        return _impl


def _authorize(allowed: SubroutineFnWrapper) -> Callable[..., HandlerFunc]:
    args = allowed.subroutine.expected_arg_types

    if len(args) != 1 or args[0] is not Expr:
        raise TealInputError(
            "Expected a single expression argument to authorize function"
        )

    if allowed.type_of() != TealType.uint64:
        raise TealTypeError(allowed.type_of(), TealType.uint64)

    def _decorate(fn: HandlerFunc) -> HandlerFunc:
        @wraps(fn)
        def _impl(*args, **kwargs) -> Expr:
            return Seq(
                Assert(allowed(Txn.sender()), comment="unauthorized"),
                fn(*args, **kwargs),
            )

        return _impl

    return _decorate


def _readonly(fn: HandlerFunc) -> HandlerFunc:
    set_handler_config(fn, read_only=True)
    return fn


def _on_complete(mc: MethodConfig) -> Callable[..., HandlerFunc]:
    def _impl(fn: HandlerFunc) -> HandlerFunc:
        set_handler_config(fn, method_config=mc)
        return fn

    return _impl


def _replace_structs(fn: HandlerFunc) -> HandlerFunc:
    sig = signature(fn)
    params = sig.parameters.copy()

    replaced = {}
    annotations = get_annotations(fn)
    for k, v in params.items():
        cls = v.annotation
        if hasattr(v.annotation, "__origin__"):
            # Generic type, not a Struct
            continue

        if issubclass(cls, abi.NamedTuple):
            params[k] = v.replace(annotation=cls().type_spec().annotation_type())
            annotations[k] = cls().type_spec().annotation_type()
            replaced[k] = cls

    if len(replaced.keys()) > 0:
        set_handler_config(fn, structs=replaced)

    newsig = sig.replace(parameters=list(params.values()))
    fn.__signature__ = newsig  # type: ignore[attr-defined]
    fn.__annotations__ = annotations

    return fn


def _capture_defaults(fn: HandlerFunc) -> HandlerFunc:
    sig = signature(fn)
    params = sig.parameters.copy()

    default_args: dict[str, DefaultArgument] = {}

    for k, v in params.items():
        match v.default:
            case Expr() | int() | str() | bytes() | FunctionType():
                default_args[k] = DefaultArgument(v.default)
                params[k] = v.replace(default=Parameter.empty)

    if len(default_args.items()) > 0:
        # Update handler config
        set_handler_config(fn, default_arguments=default_args)

        # Fix function sig/annotations
        newsig = sig.replace(parameters=list(params.values()))
        fn.__signature__ = newsig  # type: ignore[attr-defined]

    return fn


def _remove_self(fn: HandlerFunc) -> HandlerFunc:
    sig = signature(fn)
    params = sig.parameters.copy()

    if "self" in params:
        del params["self"]
        # Flag that this method did have a `self` argument
        set_handler_config(fn, referenced_self=True)
    newsig = sig.replace(parameters=list(params.values()))
    fn.__signature__ = newsig  # type: ignore[attr-defined]

    return fn


def internal(return_type_or_handler: TealType | HandlerFunc):
    """creates a subroutine to be called by logic internally

    Args:
        return_type: The type this method's returned Expression should evaluate to
    Returns:
        The wrapped subroutine
    """

    fn: Optional[HandlerFunc] = None
    return_type: Optional[TealType] = None

    if type(return_type_or_handler) is TealType:
        return_type = return_type_or_handler
    else:
        fn = cast(HandlerFunc, return_type_or_handler)

    def _impl(fn: HandlerFunc):

        if return_type is not None:
            set_handler_config(fn, subroutine=Subroutine(return_type, name=fn.__name__))

            # Don't remove self for subroutine, it fails later on in pyteal
            # during call to _validate  with invalid signature
            sig = signature(fn)
            if "self" in sig.parameters:
                set_handler_config(fn, referenced_self=True)

        else:
            fn = _remove_self(fn)
            set_handler_config(fn, method_spec=ABIReturnSubroutine(fn).method_spec())

        return fn

    if fn is not None:
        return _impl(fn)

    return _impl


def external(
    fn: HandlerFunc = None,
    /,
    *,
    authorize: SubroutineFnWrapper = None,
    method_config: MethodConfig = None,
    read_only: bool = False,
) -> HandlerFunc:

    """
    Add the method decorated to be handled as an ABI method for the Application

    Args:
        fn: The function being wrapped.
        authorize: a subroutine with input of ``Txn.sender()`` and output uint64 interpreted as allowed if the output>0.
        method_config:  A subroutine that should take a single argument (Txn.sender()) and evaluate to 1/0 depending on the app call transaction sender.
        read_only: Mark a method as callable with no fee using dryrun or simulate

    Returns:
        The original method with additional elements set in its  :code:`__handler_config__` attribute
    """

    def _impl(fn: HandlerFunc):
        fn = _remove_self(fn)
        fn = _capture_defaults(fn)
        fn = _replace_structs(fn)

        if authorize is not None:
            fn = _authorize(authorize)(fn)
        if method_config is not None:
            fn = _on_complete(method_config)(fn)
        if read_only:
            fn = _readonly(fn)

        set_handler_config(fn, method_spec=ABIReturnSubroutine(fn).method_spec())

        return fn

    if fn is None:
        return _impl

    return _impl(fn)


def bare_external(
    no_op: CallConfig = None,
    opt_in: CallConfig = None,
    clear_state: CallConfig = None,
    delete_application: CallConfig = None,
    update_application: CallConfig = None,
    close_out: CallConfig = None,
) -> Callable[..., HandlerFunc]:
    """Add method to be handled by specific bare :code:`OnComplete` actions.

    Args:
        no_op: CallConfig to handle a `NoOp`
        opt_in: CallConfig to handle an `OptIn`
        clear_state: CallConfig to handle a `ClearState`
        delete_application: CallConfig to handle a `DeleteApplication`
        update_application: CallConfig to handle a `UpdateApplication`
        close_out: CallConfig to handle a `CloseOut`

    Returns:
        The original method with changes made to its signature and attributes set in its `__handler_config__`

    """

    def _impl(fun: HandlerFunc) -> HandlerFunc:
        fun = _remove_self(fun)
        fn = Subroutine(TealType.none)(fun)
        bca = BareCallActions(
            no_op=OnCompleteAction(action=fn, call_config=no_op)
            if no_op is not None
            else OnCompleteAction.never(),
            delete_application=OnCompleteAction(
                action=fn, call_config=delete_application
            )
            if delete_application is not None
            else OnCompleteAction.never(),
            update_application=OnCompleteAction(
                action=fn, call_config=update_application
            )
            if update_application is not None
            else OnCompleteAction.never(),
            opt_in=OnCompleteAction(action=fn, call_config=opt_in)
            if opt_in is not None
            else OnCompleteAction.never(),
            close_out=OnCompleteAction(action=fn, call_config=close_out)
            if close_out is not None
            else OnCompleteAction.never(),
            clear_state=OnCompleteAction(action=fn, call_config=clear_state)
            if clear_state is not None
            else OnCompleteAction.never(),
        )

        set_handler_config(fun, bare_method=bca)

        return fun

    return _impl


def is_bare(fn: HandlerFunc) -> bool:
    sig = signature(fn)
    return len(sig.parameters) == 0 or (
        len(sig.parameters) == 1 and "self" in sig.parameters
    )


def create(
    fn: HandlerFunc = None,
    /,
    *,
    authorize: SubroutineFnWrapper = None,
    method_config: Optional[MethodConfig] = None,
):
    """set method to be handled by an application call with its :code:`OnComplete` set to :code:`NoOp` call and ApplicationId == 0

    Args:
        fn: The method to be wrapped.
    Returns:
        The original method with changes made to its signature and attributes set in its `__handler_config__`
    """

    mconfig = (
        {"no_op": CallConfig.CREATE} if method_config is None else asdict(method_config)
    )

    if not all(
        [cc == CallConfig.CREATE or cc == CallConfig.ALL for cc in mconfig.values()]
    ):
        raise TealInputError(
            "method_config for create may not have non create call configs"
        )

    def _impl(fn: HandlerFunc):
        if is_bare(fn):
            if authorize is not None:
                fn = _authorize(authorize)(fn)
            return bare_external(**mconfig)(fn)
        else:
            return external(method_config=MethodConfig(**mconfig), authorize=authorize)(
                fn
            )

    if fn is None:
        return _impl

    return _impl(fn)


def delete(fn: HandlerFunc = None, /, *, authorize: SubroutineFnWrapper = None):
    """set method to be handled by an application call with its :code:`OnComplete` set to :code:`DeleteApplication` call

    Args:
        fn: The method to be wrapped.
        authorize: a subroutine with input of ``Txn.sender()`` and output uint64 interpreted as allowed if the output>0.
    Returns:
        The original method with changes made to its signature and attributes set in its `__handler_config__`
    """

    def _impl(fn: HandlerFunc):
        if is_bare(fn):
            if authorize is not None:
                fn = _authorize(authorize)(fn)
            return bare_external(delete_application=CallConfig.CALL)(fn)
        else:
            return external(
                method_config=MethodConfig(delete_application=CallConfig.CALL),
                authorize=authorize,
            )(fn)

    if fn is None:
        return _impl

    return _impl(fn)


def update(fn: HandlerFunc = None, /, *, authorize: SubroutineFnWrapper = None):
    """set method to be handled by an application call with its :code:`OnComplete` set to :code:`UpdateApplication` call

    Args:
        fn: The method to be wrapped.
        authorize: a subroutine with input of ``Txn.sender()`` and output uint64 interpreted as allowed if the output>0.
    Returns:
        The original method with changes made to its signature and attributes set in its `__handler_config__`
    """

    # If fn has abi args, call external
    # else bare

    def _impl(fn: HandlerFunc):
        if is_bare(fn):
            if authorize is not None:
                fn = _authorize(authorize)(fn)
            return bare_external(update_application=CallConfig.CALL)(fn)
        else:
            return external(
                method_config=MethodConfig(update_application=CallConfig.CALL),
                authorize=authorize,
            )(fn)

    if fn is None:
        return _impl

    return _impl(fn)


def opt_in(fn: HandlerFunc = None, /, *, authorize: SubroutineFnWrapper = None):
    """set method to be handled by an application call with its :code:`OnComplete` set to :code:`OptIn` call

    Args:
        fn: The method to be wrapped.
        authorize: a subroutine with input of ``Txn.sender()`` and output uint64 interpreted as allowed if the output>0.
    Returns:
        The original method with changes made to its signature and attributes set in its `__handler_config__`
    """

    def _impl(fn: HandlerFunc):
        if is_bare(fn):
            if authorize is not None:
                fn = _authorize(authorize)(fn)
            return bare_external(opt_in=CallConfig.CALL)(fn)
        else:
            return external(
                method_config=MethodConfig(opt_in=CallConfig.CALL), authorize=authorize
            )(fn)

    if fn is None:
        return _impl

    return _impl(fn)


def clear_state(fn: HandlerFunc = None, /, *, authorize: SubroutineFnWrapper = None):
    """set method to be handled by an application call with its :code:`OnComplete` set to :code:`ClearState` call

    Args:
        fn: The method to be wrapped.
        authorize: a subroutine with input of ``Txn.sender()`` and output uint64 interpreted as allowed if the output>0.
    Returns:
        The original method with changes made to its signature and attributes set in its `__handler_config__`
    """

    def _impl(fn: HandlerFunc):
        if is_bare(fn):
            if authorize is not None:
                fn = _authorize(authorize)(fn)
            return bare_external(clear_state=CallConfig.CALL)(fn)
        else:
            return external(
                method_config=MethodConfig(clear_state=CallConfig.CALL),
                authorize=authorize,
            )(fn)

    if fn is None:
        return _impl

    return _impl(fn)


def close_out(fn: HandlerFunc = None, /, *, authorize: SubroutineFnWrapper = None):
    """set method to be handled by an application call with its :code:`OnComplete` set to :code:`CloseOut` call

    Args:
        fn: The method to be wrapped.
        authorize: a subroutine with input of ``Txn.sender()`` and output uint64 interpreted as allowed if the output>0.
    Returns:
        The original method with changes made to its signature and attributes set in its `__handler_config__`
    """

    def _impl(fn: HandlerFunc):
        if is_bare(fn):
            if authorize is not None:
                fn = _authorize(authorize)(fn)
            return bare_external(close_out=CallConfig.CALL)(fn)
        else:
            return external(
                method_config=MethodConfig(close_out=CallConfig.CALL),
                authorize=authorize,
            )(fn)

    if fn is None:
        return _impl

    return _impl(fn)


def no_op(fn: HandlerFunc = None, /, *, authorize: SubroutineFnWrapper = None):
    """set method to be handled by an application call with its :code:`OnComplete` set to :code:`NoOp` call

    Args:
        fn: The method to be wrapped.
        authorize: a subroutine with input of ``Txn.sender()`` and output uint64 interpreted as allowed if the output>0.
    Returns:
        The original method with changes made to its signature and attributes set in its `__handler_config__`
    """

    def _impl(fn: HandlerFunc):
        if is_bare(fn):
            if authorize is not None:
                fn = _authorize(authorize)(fn)
            return bare_external(no_op=CallConfig.CALL)(fn)
        else:
            return external(
                method_config=MethodConfig(no_op=CallConfig.CALL), authorize=authorize
            )(fn)

    if fn is None:
        return _impl

    return _impl(fn)



import pyteal as pt
from beaker.testing.unit_testing_helpers import UnitTestingApp, assert_output

from beaker.lib.iter import iterate


def test_iterate():
    ut = UnitTestingApp(
        pt.Seq(
            (buff := pt.ScratchVar()).store(pt.Bytes("")),
            iterate(buff.store(pt.Concat(buff.load(), pt.Bytes("a"))), pt.Int(10)),
            buff.load(),
        )
    )

    output = [list(b"a" * 10)]
    assert_output(ut, [], output)


def test_iterate_with_closure():
    i = pt.ScratchVar()
    buff = pt.ScratchVar()

    @pt.Subroutine(pt.TealType.none)
    def concat_thing():
        return buff.store(pt.Concat(buff.load(), pt.Itob(i.load())))

    ut = UnitTestingApp(
        pt.Seq(
            buff.store(pt.Bytes("")),
            iterate(concat_thing(), pt.Int(10), i),
            buff.load(),
        )
    )

    output = [list(b"".join([x.to_bytes(8, "big") for x in range(10)]))]

    assert_output(ut, [], output)



from .application import Application, get_method_spec
from .state import (
    AccountState,
    ApplicationState,
    ReservedApplicationStateValue,
    ReservedAccountStateValue,
    ApplicationStateValue,
    AccountStateValue,
    AccountStateBlob,
    ApplicationStateBlob,
)
from .decorators import (
    Authorize,
    external,
    internal,
    bare_external,
    create,
    no_op,
    update,
    delete,
    opt_in,
    close_out,
    clear_state,
)
from .logic_signature import LogicSignature, TemplateVariable
from .precompile import Precompile

from . import client
from . import sandbox
from . import consts
from . import lib
from . import testing



from typing import Any
from algosdk.v2client.algod import AlgodClient


def get_balances(client: AlgodClient, accts: list[str]) -> dict[str, dict[int, int]]:
    """get the balances for all the accounts in the list passed"""
    return {acct: balances(client.account_info(acct)) for acct in accts}


def get_deltas(
    acct_balances_before: dict[str, dict[int, int]],
    acct_balances_after: dict[str, dict[int, int]],
) -> dict[str, dict[int, int]]:
    """get the difference between the balances before and after some event"""
    return {
        acct: balance_delta(acct_balances_before[acct], acct_balances_after[acct])
        for acct in acct_balances_after.keys()
    }


def balances(acct_info: dict[str, Any]) -> dict[int, int]:
    """organize the balances into a dictionary of id=>amount

    Note:
        0 asset id is algos

    """
    # Init with 0 for algos
    b: dict[int, int] = {0: acct_info["amount"]}
    if "assets" in acct_info:
        for asset in acct_info["assets"]:
            b[asset["asset-id"]] = asset["amount"]

    return b


def balance_delta(
    balance_before: dict[int, int], balance_after: dict[int, int]
) -> dict[int, int]:
    """take the difference between balance after and before"""

    all_ids = list(set(list(balance_before.keys()) + list(balance_after.keys())))

    return {
        aid: balance_after.get(aid, 0) - balance_before.get(aid, 0) for aid in all_ids
    }



from dataclasses import dataclass, field
from typing import Optional
from pyteal import (
    Seq,
    Bytes,
    Expr,
    Addr,
    ScratchVar,
    TealType,
    TealTypeError,
    TealInputError,
    Int,
    Concat,
    Len,
    Substring,
    Suffix,
    Subroutine,
    Sha512_256,
)
from algosdk.source_map import SourceMap
from algosdk.future.transaction import LogicSigAccount
from algosdk.atomic_transaction_composer import LogicSigTransactionSigner
from beaker.consts import PROGRAM_DOMAIN_SEPARATOR
from beaker.lib.strings import encode_uvarint


#: The prefix for template variables that should be substituted
TMPL_PREFIX = "TMPL_"

#: The opcode that should be present just before the byte template variable
PUSH_BYTES = "pushbytes"
#: The opcode that should be present just before the uint64 template variable
PUSH_INT = "pushint"

#: The zero value for byte type
ZERO_BYTES = '""'
#: The zero value for uint64 type
ZERO_INT = "0"


@dataclass
class PrecompileTemplateValue:
    #: The name of the template variable
    name: str = field(kw_only=True)
    #: Whether or not this variable is bytes (if false, its uint64)
    is_bytes: bool = field(kw_only=True)
    #: The line number in the source TEAL this variable is present
    line: int = field(kw_only=True)
    #: The pc of the variable in the assembled bytecode
    pc: int = 0


class Precompile:
    """
    Precompile allows a contract to signal that some LogicSignature should be
    fully compiled prior to trying to construct the approval and clear programs.
    """

    def __init__(self, teal_src: str | None):

        if teal_src is None:
            raise TealInputError("teal_src cannot be None")

        self.program = teal_src

        self.binary: Optional[bytes] = None
        self.program_hash: Optional[str] = None
        self.map: Optional[SourceMap] = None

        self.template_values: list[PrecompileTemplateValue] = []

        lines = self.program.splitlines()
        # Replace the teal program TMPL_* template variables with
        # the 0 value for the given type and save the list of TemplateValues
        for idx, line in enumerate(lines):
            if TMPL_PREFIX in line:
                op, name, _, _ = line.split(" ")
                tv = PrecompileTemplateValue(
                    name=name[len(TMPL_PREFIX) :], is_bytes=op == PUSH_BYTES, line=idx
                )
                lines[idx] = line.replace(name, ZERO_BYTES if tv.is_bytes else ZERO_INT)
                self.template_values.append(tv)

        self.program = "\n".join(lines)

    def _set_compiled(self, binary: bytes, program_hash: str, map: SourceMap):
        """
        Called by application_client to set the binary/addr/map for
        this precompile.
        """
        self.binary = binary
        self.program_hash = program_hash
        self.map = map

        self.binary_bytes = Bytes(binary)

        # Denote the pc for each template value
        for tv in self.template_values:
            # +1 to acount for the pushbytes/pushint op
            tv.pc = self.map.get_pcs_for_line(tv.line)[0] + 1

    def hash(self) -> Expr:
        """
        address returns an expression for this Precompile.

        It will fail if any template_values are set.
        """

        assert self.binary is not None
        assert len(self.template_values) == 0
        if self.program_hash is None:
            raise TealInputError("No address defined for precompile")

        return Addr(self.program_hash)

    def signer(self) -> LogicSigTransactionSigner:
        """
        signer returns a LogicSigTransactionSigner to be used with
        an ApplicationClient or AtomicTransactionComposer.

        It should only be used for non templated Precompiles.
        """
        return LogicSigTransactionSigner(LogicSigAccount(self.binary))

    def populate_template(self, *args) -> bytes:
        """
        populate_template returns the bytes resulting from patching the set of
        arguments passed into the blank binary

        The args passed should be of the same type and in the same order as the
        template values declared.
        """

        assert self.binary is not None
        assert len(self.template_values) > 0
        assert len(args) == len(self.template_values)

        # Get a copy of the binary so we can work on it in place
        populated_binary = list(self.binary).copy()
        # Any time we add bytes, we need to update the offset so the rest
        # of the pc values can be updated to account for the difference
        offset = 0
        for idx, tv in enumerate(self.template_values):
            arg: str | bytes | int = args[idx]

            if tv.is_bytes:
                if type(arg) is int:
                    raise TealTypeError(type(arg), bytes | str)

                if type(arg) is str:
                    arg = arg.encode("utf-8")

                assert type(arg) is bytes

                # Bytes are encoded as uvarint(len(bytes)) + bytes
                curr_val = py_encode_uvarint(len(arg)) + arg
            else:
                if type(arg) is not int:
                    raise TealTypeError(type(arg), int)
                # Ints are just the uvarint encoded number
                curr_val = py_encode_uvarint(arg)

            # update the working buffer to include the new value, replacing the current 0 value
            populated_binary[tv.pc + offset : tv.pc + offset + 1] = curr_val

            # update the offset with the length(value) - 1 to account for the existing 0 value
            # and help keep track of how to shift the pc later
            offset += len(curr_val) - 1

        return bytes(populated_binary)

    def template_signer(self, *args) -> LogicSigTransactionSigner:
        # TODO: assert its being called with an lsig
        return LogicSigTransactionSigner(LogicSigAccount(self.populate_template(*args)))

    def populate_template_expr(self, *args: Expr) -> Expr:
        """
        populate_template_expr returns the Expr that will patch a
        blank binary given a set of arguments.

        It is called by ``template_address`` to return a Expr that
        can be used to compare with a sender given some arguments.
        """

        # To understand how this works, first look at the pure python one above
        # it should produce an identical output in terms of populated binary.
        # This function just reproduces the same effects in pyteal

        assert self.binary_bytes is not None
        assert len(self.template_values)
        assert len(args) == len(self.template_values)

        populate_program: list[Expr] = [
            (last_pos := ScratchVar(TealType.uint64)).store(Int(0)),
            (offset := ScratchVar(TealType.uint64)).store(Int(0)),
            (curr_val := ScratchVar(TealType.bytes)).store(Bytes("")),
            (buff := ScratchVar(TealType.bytes)).store(Bytes("")),
        ]

        for idx, tv in enumerate(self.template_values):
            # Add expressions to encode the values and insert them into the working buffer
            populate_program += [
                curr_val.store(Concat(encode_uvarint(Len(args[idx])), args[idx]))
                if tv.is_bytes
                else curr_val.store(encode_uvarint(args[idx])),
                buff.store(
                    Concat(
                        buff.load(),
                        Substring(
                            self.binary_bytes,
                            last_pos.load(),
                            Int(tv.pc),
                        ),
                        curr_val.load(),
                    )
                ),
                offset.store(offset.load() + Len(curr_val.load()) - Int(1)),
                last_pos.store(Int(tv.pc) + Int(1)),
            ]

        # append the bytes from the last template variable to the end
        populate_program += [
            buff.store(Concat(buff.load(), Suffix(self.binary_bytes, last_pos.load()))),
            buff.load(),
        ]

        @Subroutine(TealType.bytes)
        def populate_template_program():
            return Seq(*populate_program)

        return populate_template_program()

    def template_hash(self, *args) -> Expr:
        """
        returns an expression that will generate the expected
        hash given some set of values that should be included in the logic itself
        """
        return Sha512_256(
            Concat(Bytes(PROGRAM_DOMAIN_SEPARATOR), self.populate_template_expr(*args))
        )


def py_encode_uvarint(integer: int) -> bytes:
    """Encodes an integer as an uvarint.
    :param integer: the integer to encode
    :return: bytes containing the integer encoded as an uvarint
    """

    def to_byte(integer: int) -> int:
        return integer & 0b1111_1111

    buffer: bytearray = bytearray()

    while integer >= 0b1000_0000:
        buffer.append(to_byte(integer) | 0b1000_0000)
        integer >>= 7

    buffer.append(to_byte(integer))

    return bytes(buffer)



from pyteal import Expr, For, Int, ScratchVar, Subroutine, TealType


def iterate(sub: Expr, n: Int, i: ScratchVar = ScratchVar()) -> Expr:
    """Iterate provides a convenience method for calling a method n times

    Args:
        sub: A PyTEAL Expr to call, should not return anything
        n: The number of times to call the expression
        i: (Optional) A ScratchVar to use for iteration, passed if the caller wants to access the iterator

    Returns:
        A Subroutine expression to be passed directly into an Expr tree
    """

    @Subroutine(TealType.none)
    def _impl() -> Expr:
        init = i.store(Int(0))
        cond = i.load() < n
        iter = i.store(i.load() + Int(1))
        return For(init, cond, iter).Do(sub)

    return _impl()



import pytest
import pyteal as pt
from beaker.decorators import internal
from beaker.logic_signature import LogicSignature, TemplateVariable


def test_simple_logic_signature():
    class Lsig(LogicSignature):
        pass

    lsig = Lsig()

    assert len(lsig.template_variables) == 0
    assert len(lsig.methods) == 0
    assert len(lsig.attrs.keys()) == 1
    assert len(lsig.program) > 0

    assert "evaluate" in lsig.attrs

    assert lsig.evaluate() == pt.Reject()

    # Cant call it from static context
    with pytest.raises(TypeError):
        Lsig.evaluate()


def test_evaluate_logic_signature():
    class Lsig(LogicSignature):
        def evaluate(self):
            return pt.Approve()

    lsig = Lsig()

    assert len(lsig.template_variables) == 0
    assert len(lsig.methods) == 0
    assert len(lsig.attrs.keys()) == 1
    assert len(lsig.program) > 0

    assert "evaluate" in lsig.attrs

    assert lsig.evaluate() == pt.Approve()

    # Cant call it from static context
    with pytest.raises(TypeError):
        Lsig.evaluate()


def test_handler_logic_signature():
    class Lsig(LogicSignature):
        def evaluate(self):
            return pt.Seq(
                (s := pt.abi.String()).decode(pt.Txn.application_args[1]),
                pt.Assert(self.checked(s)),
                pt.Int(1),
            )

        @internal(pt.TealType.uint64)
        def checked(self, s: pt.abi.String):
            return pt.Len(s.get()) > pt.Int(0)

    lsig = Lsig()

    assert len(lsig.template_variables) == 0
    assert len(lsig.methods) == 0
    assert len(lsig.attrs.keys()) == 2
    assert len(lsig.program) > 0

    assert "evaluate" in lsig.attrs

    # Should not fail
    lsig.evaluate()
    lsig.checked(pt.abi.String())

    # Cant call it from static context
    with pytest.raises(TypeError):
        Lsig.evaluate()

    with pytest.raises(TypeError):
        Lsig.checked(pt.abi.String())


def test_templated_logic_signature():
    class Lsig(LogicSignature):
        pubkey = TemplateVariable(pt.TealType.bytes)

        def evaluate(self):
            return pt.Seq(
                pt.Assert(pt.Len(self.pubkey)),
                pt.Int(1),
            )

    lsig = Lsig()

    assert len(lsig.template_variables) == 1
    assert len(lsig.methods) == 0
    assert len(lsig.attrs.keys()) == 2
    assert len(lsig.program) > 0

    assert "evaluate" in lsig.attrs
    assert "pubkey" in lsig.attrs

    assert lsig.pubkey.get_name() == "TMPL_PUBKEY"

    actual = lsig.pubkey._init_expr()
    expected = pt.ScratchStore(pt.Int(1), pt.Tmpl.Bytes("TMPL_PUBKEY"))

    with pt.TealComponent.Context.ignoreScratchSlotEquality():
        assert actual == expected

    # Should not fail
    lsig.evaluate()

    # Cant call it from static context
    with pytest.raises(TypeError):
        Lsig.evaluate()


def test_different_methods_logic_signature():
    class Lsig(LogicSignature):
        def evaluate(self):
            return pt.Seq(
                (s := pt.abi.String()).decode(pt.Txn.application_args[1]),
                self.abi_tester(s, output=(o := pt.abi.Uint64())),
                pt.Assert(self.internal_tester(o.get(), pt.Len(s.get()))),
                (sv := pt.ScratchVar()).store(pt.Int(1)),
                pt.Assert(self.internal_scratch_tester(sv, o.get())),
                pt.Int(1),
            )

        @internal(None)
        def abi_tester(self, s: pt.abi.String, *, output: pt.abi.Uint64):
            return output.set(pt.Len(s.get()))

        @internal(pt.TealType.uint64)
        def internal_tester(self, x: pt.Expr, y: pt.Expr) -> pt.Expr:
            return x * y

        @internal(pt.TealType.uint64)
        def internal_scratch_tester(self, x: pt.ScratchVar, y: pt.Expr) -> pt.Expr:
            return x.load() * y

        @internal
        def no_self_abi_tester(x: pt.abi.Uint64, y: pt.abi.Uint64) -> pt.Expr:  # type: ignore
            return x.get() * y.get()

        @internal(pt.TealType.uint64)
        def no_self_internal_tester(x: pt.Expr, y: pt.Expr) -> pt.Expr:  # type: ignore
            return x * y

    lsig = Lsig()

    assert len(lsig.template_variables) == 0
    assert len(lsig.methods) == 2
    assert len(lsig.attrs.keys()) == 6
    assert len(lsig.program) > 0

    assert "evaluate" in lsig.attrs

    # Should not fail
    lsig.evaluate()
    lsig.abi_tester(pt.abi.String(), output=pt.abi.Uint64())
    lsig.internal_tester(pt.Int(1), pt.Int(1))
    lsig.internal_scratch_tester(pt.ScratchVar(), pt.Int(1))
    Lsig.no_self_abi_tester(pt.abi.Uint64(), pt.abi.Uint64())

    lsig.no_self_internal_tester(pt.Int(1), pt.Int(1))
    Lsig.no_self_internal_tester(pt.Int(1), pt.Int(1))

    # Cant call it from bound context
    with pytest.raises(TypeError):
        lsig.no_self_abi_tester(pt.abi.Uint64(), pt.abi.Uint64())



import pyteal as pt

from beaker.testing import UnitTestingApp, assert_output
from beaker.lib.inline import InlineAssembly


def test_inline_assembly():
    get_uint8 = """
extract 7 1
"""
    ut = UnitTestingApp(
        InlineAssembly(get_uint8, pt.Itob(pt.Int(255)), type=pt.TealType.bytes),
    )

    expected = [255]
    assert_output(ut, [], [expected])



import math

from pyteal import (
    BitLen,
    BytesAdd,
    BytesDiv,
    BytesMinus,
    BytesMul,
    Btoi,
    Concat,
    Exp,
    Expr,
    ExtractUint64,
    If,
    Int,
    Itob,
    Len,
    Not,
    Return,
    ScratchSlot,
    Seq,
    Subroutine,
    TealType,
)

from ..inline.inline_asm import InlineAssembly

_scale = 1000000
_log2_10 = math.log2(10)
_log2_e = math.log2(math.e)

_max_uint = (2**64) - 1
_half_uint = (2**32) - 1

log2_10 = Int(int(_log2_10 * _scale))
log2_e = Int(int(_log2_e * _scale))
scale = Int(_scale)

max_uint = Int(_max_uint)
half_uint = Int(_half_uint)


@Subroutine(TealType.uint64)
def odd(x):
    """odd returns 1 if x is odd

    Args:
        x: uint64 to evaluate

    Returns:
        uint64 1 or 0 where 1 is true and 0 is false

    """
    return x & Int(1)


@Subroutine(TealType.uint64)
def even(x):
    """even returns 1 if x is even

    Args:
        x: uint64 to evaluate

    Returns:
        uint64 1 or 0 where 1 is true and 0 is false


    """
    return Not(odd(x))


@Subroutine(TealType.uint64)
def max(a, b) -> Expr:
    """max returns the max of 2 integers"""
    return If(a > b, a, b)


@Subroutine(TealType.uint64)
def min(a, b) -> Expr:
    """min returns the min of 2 integers"""
    return If(a < b, a, b)


@Subroutine(TealType.uint64)
def saturate(n, upper_limit, lower_limit) -> Expr:
    """Produces an output that is the value of n bounded to the upper and lower
    saturation values. The upper and lower limits are specified by the
    parameters upper_limit and lower_limit."""
    return (
        If(n >= upper_limit)
        .Then(Return(upper_limit))
        .ElseIf(n <= lower_limit)
        .Then(Return(lower_limit))
        .Else(Return(n))
    )


@Subroutine(TealType.uint64)
def div_ceil(a, b) -> Expr:
    """Returns the result of division rounded up to the next integer

    Args:
        a: uint64 numerator for the operation
        b: uint64 denominator for the operation

    Returns:
        uint64 result of a truncated division + 1

    """
    q = a / b
    return If(a % b > Int(0), q + Int(1), q)


@Subroutine(TealType.uint64)
def pow10(x) -> Expr:
    """
    Returns 10^x, useful for things like total supply of an asset

    """
    return Exp(Int(10), x)


@Subroutine(TealType.uint64)
def factorial(x):
    """factorial returns x! = x * x-1 * x-2 * ...,
    for a 64bit integer, the max possible value is maxes out at 20

    Args:
        x: uint64 to call evaluate

    Returns:
        uint64 representing the factorial of the argument passed

    """
    return If(x == Int(1), x, x * factorial(x - Int(1)))


@Subroutine(TealType.bytes)
def wide_factorial(x):
    """factorial returns x! = x * x-1 * x-2 * ...,

    Args:
        x: bytes to evaluate as an integer

    Returns:
        bytes representing the integer that is the result of the factorial applied on the argument passed

    """
    return If(
        BitLen(x) == Int(1), x, BytesMul(x, wide_factorial(BytesMinus(x, Itob(Int(1)))))
    )


@Subroutine(TealType.bytes)
def wide_power(x, n):
    """wide_power returns the result of x^n evaluated using expw and combining the hi/low uint64s into a byte string

    Args:
        x: uint64 base for evaluation
        n: uint64 power to apply as exponent

    Returns:
        bytes representing the high and low bytes of a wide power evaluation

    """
    return Seq(InlineAssembly("expw", x, n), stack_to_wide())


def exponential(x, n):
    """exponential approximates e**x for n iterations

    TODO: currently this is scaled to 1000 first then scaled back. A better implementation should include the use of ufixed in abi types

    Args:
        x: The exponent to apply
        n: The number of iterations, more is better appx but costs ops

    Returns:
        uint64 that is the result of raising e**x

    """
    _scale = Itob(Int(1000))

    @Subroutine(TealType.bytes)
    def _impl(x, f, n):
        return If(
            n == Int(1),
            BytesAdd(_scale, BytesMul(x, _scale)),
            BytesAdd(
                _impl(x, BytesDiv(f, Itob(n)), n - Int(1)),
                BytesDiv(BytesMul(_scale, wide_power(bytes_to_int(x), n)), f),
            ),
        )

    return bytes_to_int(BytesDiv(_impl(Itob(x), wide_factorial(Itob(n)), n), _scale))


# @Subroutine(TealType.uint64)
# def log2(x):
#    """log2 is currently an alias for BitLen
#
#    TODO: implement with ufixed
#
#    """
#    return BitLen(x)  # Only returns integral component
#
#
# @Subroutine(TealType.uint64)
# def ln(x):
#    """Returns natural log of x for integer passed
#
#    This is heavily truncated since log2 does not return the fractional component yet
#
#    TODO: implement with ufixed
#
#    Args:
#        x: uint64 on which to take the natural log
#
#    Returns:
#        uint64 as the natural log of the value passed.
#    """
#    return (log2(x) * scale) / log2_e
#
#
# @Subroutine(TealType.uint64)
# def log10(x):
#    """Returns log base 10 of the integer passed
#
#    uses log10(x) = log2(x)/log2(10) identity
#
#    TODO: implement with ufixed
#
#    Args:
#        x: uint64  on which to take the log base 10
#
#    Returns:
#        uint64 as the log10 of the value passed
#
#    """
#    return (log2(x) * scale) / log2_10


@Subroutine(TealType.uint64)
def bytes_to_int(x):
    return If(Len(x) < Int(8), Btoi(x), ExtractUint64(x, Len(x) - Int(8)))


@Subroutine(TealType.bytes)
def stack_to_wide():
    """stack_to_wide returns the combination of the high and low integers returned from a wide math operation as bytes"""
    h = ScratchSlot()
    l = ScratchSlot()
    return Seq(
        l.store(),
        h.store(),  # Take the low and high ints off the stack and combine them
        Concat(Itob(h.load()), Itob(l.load())),
    )



from beaker.client.api_providers import (
    Network,
    Sandbox,
    AlgoNode,
    AlgoExplorer,
    PureStake,
)


def test_sandbox():
    sb = Sandbox(Network.SandNet)
    sb.algod().suggested_params()
    # sandbox indexer off
    # sb.indexer().health()


def test_algonode():
    for network in Network:
        if network == Network.SandNet:
            continue
        print(f"trying {network}")
        an = AlgoNode(network)
        an.algod().suggested_params()
        an.indexer().health()


def test_algoexplorer():
    for network in Network:
        if network == Network.SandNet:
            continue

        print(f"trying {network}")
        ae = AlgoExplorer(network)
        ae.algod().suggested_params()
        ae.indexer().health()


def test_purestake():
    return

    api_key = "TODO"
    for network in Network:
        if network == Network.SandNet:
            continue

        print(f"trying {network}")
        ae = PureStake(network)
        ae.algod(api_key).suggested_params()
        ae.indexer(api_key).health()



import base64
from inspect import getattr_static
from typing import Final, Any, cast, Optional
from algosdk.abi import Method
from pyteal import (
    SubroutineFnWrapper,
    TealInputError,
    Txn,
    MAX_TEAL_VERSION,
    ABIReturnSubroutine,
    BareCallActions,
    Expr,
    Global,
    OnCompleteAction,
    OptimizeOptions,
    Router,
    Bytes,
    Approve,
)

from beaker.decorators import (
    get_handler_config,
    MethodHints,
    MethodConfig,
    create,
)

from beaker.state import (
    AccountState,
    AccountStateBlob,
    ApplicationStateBlob,
    ApplicationState,
    ReservedAccountStateValue,
    AccountStateValue,
    ApplicationStateValue,
    ReservedApplicationStateValue,
)
from beaker.errors import BareOverwriteError
from beaker.precompile import Precompile


def get_method_spec(fn) -> Method:
    hc = get_handler_config(fn)
    if hc.method_spec is None:
        raise Exception("Expected argument to be an ABI method")
    return hc.method_spec


def get_method_signature(fn) -> str:
    return get_method_spec(fn).get_signature()


def get_method_selector(fn) -> bytes:
    return get_method_spec(fn).get_selector()


class Application:
    """Application contains logic to detect State Variables, Bare methods
    ABI Methods and internal subroutines.

    It should be subclassed to provide basic behavior to a custom application.
    """

    # Convenience constant fields
    address: Final[Expr] = Global.current_application_address()
    id: Final[Expr] = Global.current_application_id()

    def __init__(self, version: int = MAX_TEAL_VERSION):
        """Initialize the Application, finding all the custom attributes and initializing the Router"""
        self.teal_version = version

        # Is there a better way to get all the attrs declared in subclasses?
        self.attrs = {
            m: (getattr(self, m), getattr_static(self, m))
            for m in sorted(list(set(dir(self.__class__)) - set(dir(super()))))
            if not m.startswith("__")
        }

        # Initialize these ahead of time, may not
        # be set after init if len(precompiles)>0
        self.approval_program = None
        self.clear_program = None

        self.on_create = None
        self.on_update = None
        self.on_delete = None
        self.on_opt_in = None
        self.on_close_out = None
        self.on_clear_state = None

        self.hints: dict[str, MethodHints] = {}
        self.bare_externals: dict[str, OnCompleteAction] = {}
        self.methods: dict[str, tuple[ABIReturnSubroutine, Optional[MethodConfig]]] = {}
        self.precompiles: dict[str, Precompile] = {}

        acct_vals: dict[
            str, AccountStateValue | ReservedAccountStateValue | AccountStateBlob
        ] = {}
        app_vals: dict[
            str,
            ApplicationStateValue
            | ReservedApplicationStateValue
            | ApplicationStateBlob,
        ] = {}

        for name, (bound_attr, static_attr) in self.attrs.items():
            # Check for state vals
            match bound_attr:

                case AccountStateValue():
                    if bound_attr.key is None:
                        bound_attr.key = Bytes(name)
                    acct_vals[name] = bound_attr
                case ReservedAccountStateValue():
                    acct_vals[name] = bound_attr
                case AccountStateBlob():
                    acct_vals[name] = bound_attr

                case ApplicationStateBlob():
                    app_vals[name] = bound_attr
                case ApplicationStateValue():
                    if bound_attr.key is None:
                        bound_attr.key = Bytes(name)
                    app_vals[name] = bound_attr
                case ReservedApplicationStateValue():
                    app_vals[name] = bound_attr

                case Precompile():
                    self.precompiles[name] = bound_attr

            # Already dealt with these, move on
            if name in app_vals or name in acct_vals:
                continue

            # Check for externals and internal methods
            handler_config = get_handler_config(bound_attr)

            # Bare externals
            if handler_config.bare_method is not None:
                actions = {
                    oc: cast(OnCompleteAction, action)
                    for oc, action in handler_config.bare_method.__dict__.items()
                    if action.action is not None
                }

                for oc, action in actions.items():
                    if oc in self.bare_externals:
                        raise BareOverwriteError(oc)

                    # Swap the implementation with the bound version
                    if handler_config.referenced_self:
                        if not (
                            isinstance(action.action, SubroutineFnWrapper)
                            or isinstance(action.action, ABIReturnSubroutine)
                        ):
                            raise TealInputError(
                                f"Expected Subroutine or ABIReturnSubroutine, for {oc} got {action.action}"
                            )
                        action.action.subroutine.implementation = bound_attr

                    self.bare_externals[oc] = action

            # ABI externals
            elif handler_config.method_spec is not None:
                # Create the ABIReturnSubroutine from the static attr
                # but override the implementation with the bound version
                abi_meth = ABIReturnSubroutine(static_attr)

                if handler_config.referenced_self:
                    abi_meth.subroutine.implementation = bound_attr

                self.methods[name] = (abi_meth, handler_config.method_config)

                if handler_config.is_create():
                    if self.on_create is not None:
                        raise TealInputError("Multiple create methods specified")
                    self.on_create = static_attr

                if handler_config.is_update():
                    if self.on_update is not None:
                        raise TealInputError("Multiple update methods specified")
                    self.on_update = static_attr

                if handler_config.is_delete():
                    if self.on_delete is not None:
                        raise TealInputError("Multiple delete methods specified")
                    self.on_delete = static_attr

                if handler_config.is_opt_in():
                    if self.on_opt_in is not None:
                        raise TealInputError("Multiple opt in methods specified")
                    self.on_opt_in = static_attr

                if handler_config.is_clear_state():
                    if self.on_clear_state is not None:
                        raise TealInputError("Multiple clear state methods specified")
                    self.on_clear_state = static_attr

                if handler_config.is_close_out():
                    if self.on_close_out is not None:
                        raise TealInputError("Multiple close out methods specified")
                    self.on_close_out = static_attr

                self.methods[name] = (abi_meth, handler_config.method_config)
                self.hints[name] = handler_config.hints()

            # Internal subroutines
            elif handler_config.subroutine is not None:
                if handler_config.referenced_self:
                    setattr(self, name, handler_config.subroutine(bound_attr))
                else:
                    setattr(
                        self.__class__,
                        name,
                        handler_config.subroutine(static_attr),
                    )

        self.acct_state = AccountState(acct_vals)
        self.app_state = ApplicationState(app_vals)

        # If there are no precompiles, we can build the programs
        # with what we already have
        if len(self.precompiles) == 0:
            self.compile()

    def compile(self):
        self.router = Router(
            name=self.__class__.__name__,
            bare_calls=BareCallActions(**self.bare_externals),
            descr=self.__doc__,
        )

        # Add method externals
        for _, method_tuple in self.methods.items():
            method, method_config = method_tuple
            self.router.add_method_handler(
                method_call=method, method_config=method_config
            )

        # Compile approval and clear programs
        (
            self.approval_program,
            self.clear_program,
            self.contract,
        ) = self.router.compile_program(
            version=self.teal_version,
            assemble_constants=True,
            optimize=OptimizeOptions(scratch_slots=True),
        )

    def application_spec(self) -> dict[str, Any]:
        """returns a dictionary, helpful to provide to callers with information about the application specification"""

        if self.approval_program is None or self.clear_program is None:
            raise Exception(
                "approval or clear program are none, please build the programs first"
            )

        return {
            "hints": {k: v.dictify() for k, v in self.hints.items() if not v.empty()},
            "source": {
                "approval": base64.b64encode(self.approval_program.encode()).decode(
                    "utf8"
                ),
                "clear": base64.b64encode(self.clear_program.encode()).decode("utf8"),
            },
            "schema": {
                "local": self.acct_state.dictify(),
                "global": self.app_state.dictify(),
            },
            "contract": self.contract.dictify(),
        }

    def initialize_application_state(self) -> Expr:
        """
        Initialize any application state variables declared

        :return: The Expr to initialize the application state.
        :rtype: pyteal.Expr
        """
        return self.app_state.initialize()

    def initialize_account_state(self, addr=Txn.sender()) -> Expr:
        """
        Initialize any account state variables declared

        :param addr: Optional, address of account to initialize state for.
        :return: The Expr to initialize the account state.
        :rtype: pyteal.Expr
        """

        return self.acct_state.initialize(addr)

    @create
    def create(self) -> Expr:
        return Approve()

    def dump(self, directory: str = "."):
        import json
        import os

        if not os.path.exists(directory):
            os.mkdir(directory)

        with open(os.path.join(directory, "approval.teal"), "w") as f:
            if self.approval_program is None:
                raise Exception("Approval program empty")
            f.write(self.approval_program)

        with open(os.path.join(directory, "clear.teal"), "w") as f:
            if self.clear_program is None:
                raise Exception("Clear program empty")
            f.write(self.clear_program)

        with open(os.path.join(directory, "contract.json"), "w") as f:
            if self.contract is None:
                raise Exception("Contract empty")
            f.write(json.dumps(self.contract.dictify(), indent=4))

        with open(os.path.join(directory, "application.json"), "w") as f:
            f.write(json.dumps(self.application_spec(), indent=4))



from typing import Final, Callable
from pyteal import Int, Expr

#: number of microalgos in 1 Algo
algo: Final[int] = int(1e6)
#: number of microalgos in 1 MilliAlgo
milli_algo: Final[int] = int(1e3)

#: Used for runtime algo calculations `Txn.amount()==Algo`
Algo: Final[Int] = Int(algo)
#: Used for runtime algo calculations `Txn.fee()==MilliAlgo`
MilliAlgo: Final[Int] = Int(milli_algo)

#: Used for shorthand for Int(10*algo) like Algos(10)
Algos: Callable[..., Expr] = lambda v: Int(int(v * algo))
#: Used for shorthand for Int(10*milli_algo) like MilliAlgos(10)
MilliAlgos: Callable[..., Expr] = lambda v: Int(int(v * milli_algo))

#: Max number of inner transactions that may be called
MAX_INNERS = 255
#: Single app call opcode budget
APP_CALL_BUDGET = 700
#: Max possible opcode budget
MAX_OPS = MAX_INNERS * APP_CALL_BUDGET

#: Single app call budget
AppCallBudget = Int(APP_CALL_BUDGET)
#: Max app call budget possible
MaxOps = Int(MAX_OPS)


#: TRUE used as an alias for 1
TRUE: Final[Int] = Int(1)
#: FALSE used as an alias for 0
FALSE: Final[Int] = Int(0)

#: The max number of local state values that may be declared
MAX_LOCAL_STATE = 16
#: The max number of global state values that may be declared
MAX_GLOBAL_STATE = 64

#: The maximum number of args that may be included in an lsig
LSIG_MAX_ARGS = 255

#: The prefix used when hashing bytecode to produce a unique hash
PROGRAM_DOMAIN_SEPARATOR = "Program"



import pytest
import pyteal as pt
from beaker.application import Application
from beaker.decorators import external
from beaker.client import ApplicationClient
from beaker.sandbox import get_accounts, get_algod_client

from beaker.logic_signature import LogicSignature, TemplateVariable

from beaker.precompile import Precompile, py_encode_uvarint


def test_precompile():
    class App(Application):
        class Lsig(LogicSignature):
            def evaluate(self):
                return pt.Seq(pt.Assert(pt.Int(1)), pt.Int(1))

        pc = Precompile(Lsig(version=6).program)

        @external
        def check_it(self):
            return pt.Assert(pt.Txn.sender() == self.pc.hash())

    app = App()
    ac = ApplicationClient(get_algod_client(), app, signer=get_accounts().pop().signer)

    assert app.approval_program is None
    assert app.clear_program is None
    assert app.pc.program_hash is None

    ac.build()

    assert app.approval_program is not None
    assert app.clear_program is not None
    assert app.pc.program_hash is not None


TMPL_BYTE_VALS = [
    ("abc"),
    ("asdfasdfasdf"),
    (bytes(100)),
    ("的的的的的的的"),
]


@pytest.mark.parametrize("tmpl_val", TMPL_BYTE_VALS)
def test_templated_bytes(tmpl_val: str):
    class Lsig(LogicSignature):
        tv = TemplateVariable(pt.TealType.bytes)

        def evaluate(self):
            return pt.Seq(pt.Assert(pt.Len(self.tv)), pt.Int(1))

    class App(Application):
        pc = Precompile(Lsig(version=6).program)

        @external
        def check_it(self):
            return pt.Assert(
                pt.Txn.sender() == self.pc.template_hash(pt.Bytes(tmpl_val))
            )

    app = App()
    ac = ApplicationClient(get_algod_client(), app, signer=get_accounts().pop().signer)

    assert app.approval_program is None
    assert app.clear_program is None
    assert app.pc.program_hash is None

    ac.build()

    assert app.approval_program is not None
    assert app.clear_program is not None
    assert app.pc.program_hash is not None

    populated_teal = app.pc.populate_template(tmpl_val)

    vlen = len(tmpl_val)
    if type(tmpl_val) is str:
        vlen = len(tmpl_val.encode("utf-8"))

    assert len(populated_teal) == len(app.pc.binary) + vlen + (
        len(py_encode_uvarint(vlen)) - 1
    )


TMPL_INT_VALS = [(10), (1000), (int(2.9e9))]


@pytest.mark.parametrize("tmpl_val", TMPL_INT_VALS)
def test_templated_ints(tmpl_val: int):
    class Lsig(LogicSignature):
        tv = TemplateVariable(pt.TealType.uint64)

        def evaluate(self):
            return pt.Seq(pt.Assert(self.tv), pt.Int(1))

    class App(Application):
        pc = Precompile(Lsig(version=6).program)

        @external
        def check_it(self):
            return pt.Assert(pt.Txn.sender() == self.pc.template_hash(pt.Int(tmpl_val)))

    app = App()
    ac = ApplicationClient(get_algod_client(), app, signer=get_accounts().pop().signer)

    assert app.approval_program is None
    assert app.clear_program is None
    assert app.pc.program_hash is None

    ac.build()

    assert app.approval_program is not None
    assert app.clear_program is not None
    assert app.pc.program_hash is not None

    populated_teal = app.pc.populate_template(tmpl_val)

    assert len(populated_teal) == len(app.pc.binary) + (
        len(py_encode_uvarint(tmpl_val)) - 1
    )



from typing import Optional
from pyteal import (
    App,
    Bytes,
    Concat,
    Expr,
    Extract,
    For,
    GetByte,
    If,
    Int,
    Len,
    Or,
    ScratchVar,
    Seq,
    SetByte,
    Subroutine,
    Substring,
    TealType,
)

from beaker.lib.inline import InlineAssembly
from beaker.lib.storage.blob import BLOB_PAGE_SIZE, EMPTY_PAGE, Blob


class GlobalBlob(Blob):
    def __init__(self, /, *, keys: Optional[int | list[int]] = None):
        super().__init__(64, keys=keys)

    def zero(self) -> Expr:
        """
        initializes global state of an application to all zero bytes

        This allows us to be lazy later and _assume_ all the strings are the same size
        """

        @Subroutine(TealType.none)
        def zero_impl():
            zloop = """
    zero_loop:
        int 1
        -               // ["00"*page_size, key-1]
        dup2            // ["00"*page_size, key, "00"*page_size, key]
        itob            // ["00"*page_size, key, "00"*page_size, itob(key)]
        extract 7 1     // ["00"*page_size, key, "00"*page_size, itob(key)[-1]] get the last byte of the int
        swap            // ["00"*page_size, key, itob(key)[-1], "00"*page_size]
        app_global_put  // ["00"*page_size, key]  (removes top 2 elements)
        dup             // ["00"*page_size, key-1, key-1]
        bnz zero_loop   // start loop over if key-1>0
        pop
        pop             // take extra junk off the stack
        retsub
    callsub zero_loop
            """
            return InlineAssembly(zloop, EMPTY_PAGE, self.max_keys, type=TealType.none)

        return zero_impl()

    def get_byte(self, idx):
        """
        Get a single byte from global storage of an application by index
        """

        @Subroutine(TealType.uint64)
        def get_byte_impl(idx):
            return GetByte(
                App.globalGet(self._key(self._key_idx(idx))), self._offset_for_idx(idx)
            )

        return get_byte_impl(idx)

    def set_byte(self, idx, byte):
        """
        Set a single byte from global storage of an application by index
        """

        @Subroutine(TealType.none)
        def set_byte_impl(idx, byte):
            return Seq(
                (key := ScratchVar()).store(self._key(self._key_idx(idx))),
                App.globalPut(
                    key.load(),
                    SetByte(App.globalGet(key.load()), self._offset_for_idx(idx), byte),
                ),
            )

        return set_byte_impl(idx, byte)

    def read(self, bstart, bstop) -> Expr:
        """
        read bytes between bstart and bend from global storage of an application by index
        """

        @Subroutine(TealType.bytes)
        def read_impl(bstart, bstop):
            start_key = self._key_idx(bstart)
            start_offset = self._offset_for_idx(bstart)

            stop_key = self._key_idx(bstop)
            stop_offset = self._offset_for_idx(bstop)

            key = ScratchVar()
            buff = ScratchVar()

            start = ScratchVar()
            stop = ScratchVar()

            init = key.store(start_key)
            cond = key.load() <= stop_key
            incr = key.store(key.load() + Int(1))

            return Seq(
                buff.store(Bytes("")),
                For(init, cond, incr).Do(
                    Seq(
                        start.store(If(key.load() == start_key, start_offset, Int(0))),
                        stop.store(
                            If(key.load() == stop_key, stop_offset, BLOB_PAGE_SIZE)
                        ),
                        buff.store(
                            Concat(
                                buff.load(),
                                Substring(
                                    App.globalGet(self._key(key.load())),
                                    start.load(),
                                    stop.load(),
                                ),
                            )
                        ),
                    )
                ),
                buff.load(),
            )

        return read_impl(bstart, bstop)

    def write(self, bstart, buff) -> Expr:
        """
        write bytes between bstart and len(buff) to global storage of an application
        """

        @Subroutine(TealType.none)
        def write_impl(bstart, buff):
            start_key = self._key_idx(bstart)
            start_offset = self._offset_for_idx(bstart)

            stop_key = self._key_idx(bstart + Len(buff))
            stop_offset = self._offset_for_idx(bstart + Len(buff))

            key = ScratchVar()
            start = ScratchVar()
            stop = ScratchVar()
            written = ScratchVar()

            init = key.store(start_key)
            cond = key.load() <= stop_key
            incr = key.store(key.load() + Int(1))

            delta = ScratchVar()

            return Seq(
                written.store(Int(0)),
                For(init, cond, incr).Do(
                    Seq(
                        start.store(If(key.load() == start_key, start_offset, Int(0))),
                        stop.store(
                            If(key.load() == stop_key, stop_offset, BLOB_PAGE_SIZE)
                        ),
                        App.globalPut(
                            self._key(key.load()),
                            If(
                                Or(
                                    stop.load() != BLOB_PAGE_SIZE,
                                    start.load() != Int(0),
                                )
                            )  # Its a partial write
                            .Then(
                                Seq(
                                    delta.store(stop.load() - start.load()),
                                    Concat(
                                        Substring(
                                            App.globalGet(self._key(key.load())),
                                            Int(0),
                                            start.load(),
                                        ),
                                        Extract(buff, written.load(), delta.load()),
                                        Substring(
                                            App.globalGet(self._key(key.load())),
                                            stop.load(),
                                            BLOB_PAGE_SIZE,
                                        ),
                                    ),
                                )
                            )
                            .Else(
                                Seq(
                                    delta.store(BLOB_PAGE_SIZE),
                                    Extract(buff, written.load(), BLOB_PAGE_SIZE),
                                )
                            ),
                        ),
                        written.store(written.load() + delta.load()),
                    )
                ),
            )

        return write_impl(bstart, buff)
